//------------------------------------------------------ 2021-08-16 --------------
// Laminator temperature controler  (Arduino 1.8.13)
//
// Board: Arduino Nano Atmega328 16MHz
// Library:
//   TM1637 by Avishay Orpaz V1.2.0
//----------------------------- written by Hiroyuki Iizuka ---- Japan ------------
#include <TM1637Display.h>
#include <EEPROM.h>
#include <MsTimer2.h>

//----------------------------------------------------
#define SERIES_RESISTANCE   4608.5  // R1 ohm
#define DISPLAY_TEMPERATURE   96.0  // Celsius
#define TRUE_TEMPERATURE     100.0  // Celsius
//----------------------------------------------------

#define PERIOD    2000   // heater PWM period [msec]

#define TM1637_CLK    A3    // 7segment LED clock
#define TM1637_DIO    A2    // 7segment LED data

#define THERMISTOR         A0   // NTC thermistor

#define HEATER             2   // photo triac
#define HEATER_OFF 0
#define HEATER_ON  1

#define ENCODER_A         11   // rotary encoder
#define ENCODER_B          9   // rotary encoder
#define ENCODER_SW        10   // rotary encoder
#define SW_ON   0
#define SW_OFF  1
#define CW             1
#define CCW            2
#define ENCODER_STOP   0
#define ROTARY_LOCK     1
#define ROTARY_UNLOCK   0

#define MIN_TEMPERATURE  100
#define MAX_TEMPERATURE  140

#define EEPADDR_TEMPERATURE 0

TM1637Display display(TM1637_CLK, TM1637_DIO);

unsigned char Rotary_Enc_A;
unsigned char Rotary_Enc_B;
boolean Rotary_A_Lock;
boolean Rotary_B_Lock;
unsigned char Rotary_Stat;

long          target_temperature;     // target temperature
double        current_temperature;    // current temperature 
boolean       enable_sensor;
boolean       enable_encoder;
boolean       change_target_temperature;
unsigned long sensor_disable_time;
unsigned long encoder_enable_time;
unsigned char heater_pwm_duty;        // 0 to 100


//---- Rotary encoder --------
// RETURN
//    CW / CCW / ENCODER_STOP
unsigned char rotary_encoder()
{
  unsigned char temp_A;
  unsigned char temp_B;
  unsigned char result;
  
  temp_A = digitalRead( ENCODER_A );
  temp_B = digitalRead( ENCODER_B );
  
  result = ENCODER_STOP;

  if( (Rotary_Enc_A == 1) && (temp_A == 0) && (temp_B == 1) ) {
    Rotary_Stat   = CW;
    Rotary_A_Lock = true;
  }
  if( Rotary_A_Lock ) {
    if( (temp_A == 0) && (temp_B == 0) ) {
      result = Rotary_Stat;
      Rotary_A_Lock = false;
    }
  }

  if( (Rotary_Enc_B == 1) && (temp_B == 0) && (temp_A == 1) ) {
    Rotary_Stat   = CCW;
    Rotary_B_Lock = true;
  }
  if( Rotary_B_Lock ) {
    if( (temp_A == 0) && (temp_B == 0) ) {
      result = Rotary_Stat;
      Rotary_B_Lock = false;
    }
  }

  Rotary_Enc_A = temp_A;
  Rotary_Enc_B = temp_B;

  return result;
}

//--------------- calc B constant ----------------------
double calc_Bconstant( 
  double Tn, // temperature celsius
  double Rn, // NTC thermistor resistance 
  double R0) // NTC thermistor resistance @ 25 degrees_celsius
{
  return log(Rn/R0) / (1/celsius_to_kelvin(Tn) - 1/celsius_to_kelvin(25.0));
}

//--------------- calc temperature ----------------------
double get_temperature_Mitsubishi(
  double series_resistance,
  double R0, // NTC thermistor resistance @ 25 degrees_celsius
  unsigned int B_constant, // @25℃
  double t1, // temperature celsius
  double t2, // temperature celsiusca
  double t3, // temperature celsius
  double R1, // NTC thermistor resistance @ t1 degrees celsius
  double R2, // NTC thermistor resistance @ t2 degrees celsius
  double R3, // NTC thermistor resistance @ t3 degrees celsius
  double disp_temperature, // display temperature  
  double true_temperature) // true    temperatureE  
{
  double Bn,Bc1,Bc2,Bc3;
  double Tprev,Tn,T1,T2,T3;
  double C,D,E;
  double Kdisp_temp,Ktrue_temp,disp_B_const, true_B_const, E_calibration;
    
  Bc1 = calc_Bconstant( t1, R1, R0 );
  Bc2 = calc_Bconstant( t2, R2, R0 );
  Bc3 = calc_Bconstant( t3, R3, R0 );
  T1 = celsius_to_kelvin( t1 );
  T2 = celsius_to_kelvin( t2 );
  T3 = celsius_to_kelvin( t3 );
  C = ((Bc1-Bc2)*(T2-T3)-(Bc2-Bc3)*(T1-T2)) / ((T1-T2)*(T2-T3)*(T1-T3));
  D = (Bc1-Bc2-C*(T1+T2)*(T1-T2)) / (T1-T2);
  E = Bc1 - D*T1 - C*T1*T1;

  Kdisp_temp = celsius_to_kelvin( disp_temperature );
  Ktrue_temp = celsius_to_kelvin( true_temperature );

  disp_B_const = C * Kdisp_temp*Kdisp_temp + D * Kdisp_temp + E;
  true_B_const = disp_B_const * (1/Kdisp_temp - 1/celsius_to_kelvin(25.0)) / (1/Ktrue_temp - 1/celsius_to_kelvin(25.0));
  E_calibration = true_B_const - disp_B_const;
  E += E_calibration;

  Tprev = get_temperature( series_resistance, R0, B_constant, true );
  for( unsigned char i = 0 ; i < 2 ; i++ ) {
    Bn = C * Tprev*Tprev + D * Tprev + E;
    Tn = get_temperature( series_resistance, R0, Bn, false );
    if( (Tprev - 0.01 < Tn) && (Tn < Tprev + 0.01) ) {
      break;
    } else {
      Tprev = Tn;
    }
  }

  return Tn;   // Kelvin
}
  
//----------------------- temperature conversion -----------------------------
double celsius_to_kelvin( double celsius )
{
  return celsius + 273.15;
}

double kelvin_to_celsius( double kelvin )
{
  return kelvin - 273.15;
}

//----------- calc temperature -----------------------------
double get_temperature(
  unsigned int series_resistance,
  double       R0, // NTC thermistor resistance @ 25 degrees_celsius
  unsigned int B_constant,
  boolean      exec_ad )
{
  static double ntc_resistance;

  if( exec_ad ) {
    ntc_resistance = get_ntc_resistance( series_resistance );
  }
  
  return 1 / (log(ntc_resistance/R0)/B_constant + 1/celsius_to_kelvin(25.0));
}  

//---------- A/D ntc resistance ---------------
double get_ntc_resistance( unsigned int series_resistance )
{
  double ad_value;

  ad_value = 0.0;
  for( unsigned char i = 0 ; i < 100 ; i++ ) {
    ad_value += (double)analogRead( THERMISTOR );
    delay(1);
  }
  ad_value /= 100;
 
  return ( ad_value * series_resistance ) / ( 1024 - ad_value );
}

//------- EEPROM -> target_temperature  ----------------------
unsigned char get_target_temperature()
{
  unsigned char temperature;
  
  temperature = EEPROM.read( EEPADDR_TEMPERATURE );
  
  if( (temperature < MIN_TEMPERATURE) || (MAX_TEMPERATURE < temperature) ) {
    temperature = MIN_TEMPERATURE;
    write_temperature( temperature );
  }

  return temperature;
}

//------- target_temperature -> EEPROM ----------------------
void write_temperature( unsigned char temperature )
{
   EEPROM.write( EEPADDR_TEMPERATURE, temperature );
}

//---------- MsTimer2 interrupt routine --------------
void heater()
{
  static unsigned char counter = 0;

  if( 100 <= counter ) {
    counter = 0;
  }
  if( counter < heater_pwm_duty ) {
    digitalWrite( HEATER, HEATER_ON  );
  } else {     
    digitalWrite( HEATER, HEATER_OFF );
  }
  counter++;
}

//----------------------- setup -----------------------------------
void setup(){
  pinMode( THERMISTOR, INPUT );
  pinMode( ENCODER_A,  INPUT );
  pinMode( ENCODER_B,  INPUT );
  pinMode( ENCODER_SW, INPUT );
  pinMode( HEATER,     OUTPUT );

  digitalWrite( HEATER, HEATER_OFF );
  delay(500);

  display.setBrightness(7);
  display.clear();

  target_temperature = get_target_temperature();
  display.showNumberDecEx(target_temperature * 10, 0b00100000, false);
  delay(1000);
  
  Rotary_Enc_A = digitalRead( ENCODER_A );
  Rotary_Enc_B = digitalRead( ENCODER_B );
  Rotary_A_Lock = false;
  Rotary_B_Lock = false;
  
  Serial.begin(9600);
  Serial.println("-------- Laminator temperature controller -------------------");
  
  enable_sensor = true;
  enable_encoder = false;        
  MsTimer2::set(PERIOD/100, heater);  
  MsTimer2::start();

  heater_pwm_duty = 0;
  change_target_temperature = false;
}

//----------------------- loop ------------------------------------
void loop()
{
 // long encoder_value;
  unsigned char turn;
  
  if( enable_sensor ) {
    current_temperature = kelvin_to_celsius(get_temperature_Mitsubishi( SERIES_RESISTANCE, 100000, 4100, 50, 100, 150, 34470, 6092, 1569, DISPLAY_TEMPERATURE, TRUE_TEMPERATURE ));
    if( !enable_encoder ) {
      display.showNumberDecEx( current_temperature * 10, 0b00100000, false);
    }
    enable_sensor = false;
    sensor_disable_time = millis();
    Serial.print("temp=");    Serial.print( current_temperature ); Serial.print(" ");
    Serial.print("target=");  Serial.print( target_temperature );  Serial.print(" ");
    Serial.print("PWM duty=");Serial.print( heater_pwm_duty );     Serial.println(" ");
    heater_pwm_duty = max( 0, min( (  target_temperature - current_temperature ) * 15, 100 ));
  }

  if( sensor_disable_time + 1000 < millis() ) {
    enable_sensor = true;
  }
  
  if( enable_encoder ) {
    turn = rotary_encoder();
    if( turn == CW && target_temperature < MAX_TEMPERATURE ) {
      target_temperature++;
      display.showNumberDecEx( target_temperature * 10, 0b00100000, false);
      change_target_temperature = true;      
      encoder_enable_time = millis();
    }

    if( turn == CCW && MIN_TEMPERATURE < target_temperature)  {
      target_temperature--;
      display.showNumberDecEx( target_temperature * 10, 0b00100000, false);
      change_target_temperature = true;
      encoder_enable_time = millis();
    }
    sensor_disable_time = millis();
  }
  
  if( encoder_enable_time + 2000 < millis() ) {
    enable_encoder = false;  
    if( change_target_temperature ) {
      write_temperature( target_temperature );
      change_target_temperature = false;
    }
  }

  if( digitalRead(ENCODER_SW) == SW_ON ) {
    enable_encoder = true;
    encoder_enable_time = millis();
    enable_sensor = false;
    sensor_disable_time = millis();
    display.showNumberDecEx( target_temperature * 10, 0b00100000, false);
  }
}

//--------------- E N D -------------------------------------------------------------------
